/*
 * Copyright (c) 2015-2016, Texas Instruments Incorporated
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * *  Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * *  Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * *  Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/* XDCtools Header files */
#include <xdc/std.h>
#include <xdc/runtime/Assert.h>
#include <xdc/runtime/System.h>

/* TI-RTOS Header files */
#include <ti/drivers/PIN.h>
#include <ti/sysbios/BIOS.h>
#include <ti/sysbios/knl/Task.h>

/* DriverLib Header files */
#include <driverlib/flash.h>
#include <driverlib/vims.h>

/* Board Header files */
#include "Board.h"

/* Page size is 4KiB on the CC13xx/CC26xx */
#define FLASH_PAGE_SIZE         0x1000

/* Symbols are defined in the linker script. */
extern const uint32_t __FLASH_ADDR;
extern const uint32_t __FLASH_SIZE;
extern const uint32_t __FLASH_DATA_ADDR;
extern const uint32_t __FLASH_DATA_SIZE;
const uint32_t FLASH_BASE_ADDR = (uint32_t)&__FLASH_ADDR;
const uint32_t FLASH_SIZE      = (uint32_t)&__FLASH_SIZE;
const uint32_t FLASH_DATA_ADDR = (uint32_t)&__FLASH_DATA_ADDR;
const uint32_t FLASH_DATA_SIZE = (uint32_t)&__FLASH_DATA_SIZE;

/* An imagined memory structure that is kept in flash */
typedef struct
{
    uint8_t led1Enabled;
    uint8_t led2Enabled;
    uint8_t led3Enabled;
    uint8_t led4Enabled;
} ApplicationConfig;

/*
The config structure in SRAM. It is used during
run-time.
 */
ApplicationConfig configInRam;

/*
The same config structure in FLASH.
It is linked automatically into the .flashdata section and initialized
for the first use.
 */
#pragma DATA_SECTION(configInFlash, ".flashdata");
const ApplicationConfig configInFlash =
{
    .led1Enabled = 0,
    .led2Enabled = 0,
    .led3Enabled = 0,
    .led4Enabled = 0
};

/* Pin driver handles */
static PIN_Handle buttonPinHandle;
static PIN_Handle ledPinHandle;

/* Global memory storage for a PIN_Config table */
static PIN_State buttonPinState;
static PIN_State ledPinState;

/*
 * Initial LED pin configuration table
 *   - LEDs Board_LED0 is on.
 *   - LEDs Board_LED1 is off.
 */
PIN_Config ledPinTable[] = {
    Board_LED1 | PIN_GPIO_OUTPUT_EN | PIN_GPIO_HIGH | PIN_PUSHPULL | PIN_DRVSTR_MAX,
    Board_LED2 | PIN_GPIO_OUTPUT_EN | PIN_GPIO_LOW  | PIN_PUSHPULL | PIN_DRVSTR_MAX,
    Board_LED3 | PIN_GPIO_OUTPUT_EN | PIN_GPIO_LOW  | PIN_PUSHPULL | PIN_DRVSTR_MAX,
    Board_LED4 | PIN_GPIO_OUTPUT_EN | PIN_GPIO_LOW  | PIN_PUSHPULL | PIN_DRVSTR_MAX,
    PIN_TERMINATE
};

/*
 * Application button pin configuration table:
 *   - Buttons interrupts are configured to trigger on falling edge.
 */
PIN_Config buttonPinTable[] = {
    Board_KEY_UP   | PIN_INPUT_EN | PIN_PULLUP | PIN_IRQ_NEGEDGE,
    Board_KEY_LEFT | PIN_INPUT_EN | PIN_PULLUP | PIN_IRQ_NEGEDGE,
    Board_KEY_DOWN | PIN_INPUT_EN | PIN_PULLUP | PIN_IRQ_NEGEDGE,
    Board_KEY_RIGHT | PIN_INPUT_EN | PIN_PULLUP | PIN_IRQ_NEGEDGE,
    PIN_TERMINATE
};


/*
A scratch buffer in SRAM. Unfortunately it consumes 4KiB RAM on the
CC13xx because of the flash page size. It would be possible to create
it on-the-fly on stack, but then the maximum stack size has to be adjusted
accordingly.
*/
uint8_t scratchBuffer[FLASH_PAGE_SIZE];

/*
Updates nBytes data in flash memory at destAddr.

This function is a proof-of-concept implementation that demonstrates the usage of the
driverlib flash API. It must not be used in a real-application.

Existing data in the flash page is preserved.
This implementation is straight-forward and there is a lot room for optimizations.
For instance, one could check whether a write without erase is sufficient.
In general, it is a good idea to accumulate some changes in RAM first and then overwrite
everything at once instead of calling this function multiple times for the same page.

Since there is always the danger of a power-loss/reset during writing,
a ping-pong write algorithm should be used that distributes the data on two
sections.

*/
void updateMemoryInFlash(void* destAddr, const void* sourceAddr, uint16_t nBytes)
{
    /* Calculate the start address of the destination page */
    uint32_t pageStartAddr = (uint32_t)destAddr - ((uint32_t)destAddr % FLASH_PAGE_SIZE);

    /* Check overall boundaries */
    Assert_isTrue((uint32_t)destAddr >= FLASH_DATA_ADDR, 0);
    Assert_isTrue(((uint32_t)destAddr + nBytes) < (FLASH_DATA_ADDR + FLASH_DATA_SIZE), 1);

    /* Check that we do not cross a page boundary when FLASH_DATA_SIZE is more than one page */
    Assert_isTrue(((uint32_t)destAddr - pageStartAddr + nBytes) <= FLASH_PAGE_SIZE, 2);

    /*
     Make sure that the flash functions from driverlib are either located in ROM or SRAM,
     but not in FLASH. This is the normal case in TI-RTOS.
     */
    Assert_isTrue((uint32_t)&FlashProgram > (FLASH_BASE_ADDR + FLASH_SIZE), 3);

    /* Rescue the data from flash before erasing the whole page */
    memcpy(scratchBuffer, (void*)pageStartAddr, FLASH_PAGE_SIZE);

    /* Merge old and new data in the scratch buffer */
    uint32_t pageOffset = (uint32_t)destAddr - pageStartAddr;
    memcpy(scratchBuffer + pageOffset, sourceAddr, nBytes);

    /* Disable flash cache */
    VIMSModeSet(VIMS_BASE, VIMS_MODE_DISABLED);
    while (VIMSModeGet(VIMS_BASE) != VIMS_MODE_DISABLED);

    /* Erase the whole page */
    FlashSectorErase((uint32_t)pageStartAddr);

    /* Write back from the scratch buffer */
    FlashProgram(scratchBuffer, pageStartAddr, FLASH_PAGE_SIZE);

    /* Re-enable flash cache */
    VIMSModeSet(VIMS_BASE, VIMS_MODE_ENABLED);
}

void buttonCallbackFxn(PIN_Handle handle, PIN_Id pinId) {
    Bool updated = FALSE;

    /* Debounce logic, only toggle if the button is still pushed (low) */
    CPUdelay(8000*50);
    if (!PIN_getInputValue(pinId)) {
        /* Toggle LED based on the button pressed */
        switch (pinId) {
            case Board_KEY_UP:
                configInRam.led1Enabled ^= 0xFF;
                updated = TRUE;
                PIN_setOutputValue(ledPinHandle, Board_LED1, configInRam.led1Enabled);
                break;
            case Board_KEY_RIGHT:
                configInRam.led2Enabled ^= 0xFF;
                updated = TRUE;
                PIN_setOutputValue(ledPinHandle, Board_LED2, configInRam.led2Enabled);
                break;
            case Board_KEY_DOWN:
                configInRam.led3Enabled ^= 0xFF;
                updated = TRUE;
                PIN_setOutputValue(ledPinHandle, Board_LED3, configInRam.led3Enabled);
                break;
            case Board_KEY_LEFT:
                configInRam.led4Enabled ^= 0xFF;
                updated = TRUE;
                PIN_setOutputValue(ledPinHandle, Board_LED4, configInRam.led4Enabled);
                break;
        }

        if (updated == TRUE)
        {
            updateMemoryInFlash(&configInFlash, &configInRam, sizeof(ApplicationConfig));
        }
    }
}


/*
 *  ======== main ========
 */
int main(void)
{
    /* Call board init functions */
    Board_initGeneral();

    /* Initialize LED and button pins */
    ledPinHandle = PIN_open(&ledPinState, ledPinTable);
    if(!ledPinHandle) {
        System_abort("Error initializing board LED pins\n");
    }

    buttonPinHandle = PIN_open(&buttonPinState, buttonPinTable);
    if(!buttonPinHandle) {
        System_abort("Error initializing button pins\n");
    }

    /* Setup callback for button pins */
    if (PIN_registerIntCb(buttonPinHandle, &buttonCallbackFxn) != 0) {
        System_abort("Error registering button callback function");
    }

    /* Read the config from flash */
    memcpy(&configInRam, &configInFlash, sizeof(ApplicationConfig));
    PIN_setOutputValue(ledPinHandle, Board_LED1, configInRam.led1Enabled);
    PIN_setOutputValue(ledPinHandle, Board_LED2, configInRam.led2Enabled);
    PIN_setOutputValue(ledPinHandle, Board_LED3, configInRam.led3Enabled);
    PIN_setOutputValue(ledPinHandle, Board_LED4, configInRam.led4Enabled);

    /* Start BIOS */
    BIOS_start();

    return (0);
}
